﻿Namespace GeoCalc

    ' 項を表すクラス
    Public Class T

        ' 整数型の列でタプル内の値を指定する
        Public Sub New(ByVal ParamArray Values() As Long)

            ReDim val(Values.Length() - 1)
            Dim index As Long
            For index = 0 To Values.Length() - 1
                val(index) = Values(index)
            Next

        End Sub

        ' タプルの指定位置の値を取り出す
        Public Function GetC(ByVal index As Long) As Long
            Return val(index)
        End Function


        ' タプル数を返す
        Public Function GetTupleSize() As Long
            Return val.Length()
        End Function

        ' 比較演算子を定義する
        Public Shared Operator =(ByVal t1 As T, ByVal t2 As T) As Boolean

            If t1.GetTupleSize <> t2.GetTupleSize Then
                Return False
            End If
            Dim index As Long
            For index = 0 To t1.GetTupleSize - 1
                If t1.GetC(index) <> t2.GetC(index) Then
                    Return False
                End If
            Next
            Return True
        End Operator
        Public Shared Operator <>(ByVal t1 As T, ByVal t2 As T) As Boolean
            Return Not (t1 = t2)
        End Operator

        ' 項同士のAnd演算の定義
        Public Function AndT(ByVal v1 As T) As T
            If GetTupleSize() <> v1.GetTupleSize Then
                Throw New ApplicationException("項の数の違うタプルのAndを計算しました")
            End If

            Dim ans As Long()
            ReDim ans(v1.GetTupleSize() - 1)

            Dim index As Long
            For index = 0 To v1.GetTupleSize() - 1
                ans(index) = v1.GetC(index) And GetC(index)
            Next
            Return New T(ans)
        End Function

        ' 複製を作成する
        Public Function Copy() As T
            Return New T(val)
        End Function

        ' すべて0のTを作成する
        Public Function Zero() As T
            Dim ret As Long()
            ReDim ret(val.Length - 1)
            Dim index As Long
            For index = 0 To val.Length - 1
                ret(index) = 0
            Next
            Return New T(ret)
        End Function

        ' すべて１のTを作成する
        Public Function One() As T
            Dim ret As Long()
            ReDim ret(val.Length - 1)
            Dim index As Long
            For index = 0 To val.Length - 1
                ret(index) = 7
            Next
            Return New T(ret)
        End Function

        ' 特定の位置を反転する
        Public Function NegC(ByVal index As Long) As T
            Dim v1 As T
            v1 = One()
            v1.val(index) = val(index) Xor 7
            Return v1
        End Function

        '項のnot演算の定義
        Public Function NegT() As TS
            Dim ret As TS = New TS()
            Dim index As Long
            For index = 0 To GetTupleSize() - 1
                ret.AddSimple(NegC(index))
            Next
            Return ret
        End Function

        ' 文字列に変換する
        Public Overrides Function ToString() As String
            Dim ret As String
            ret = "("
            Dim index As Long
            For index = 0 To val.Length() - 1
                If index > 0 Then ret += ","
                ret += Convert.ToString(val(index) + 8, 2).Substring(1) ' "1???"という文字列直したあと，先頭の1を削除する
            Next
            Return ret + ")"
        End Function

        ' 和の公式が適用可能かをチェックする
        Public Function CanSimpleOr(ByVal t1 As T) As Boolean
            If t1.GetTupleSize() <> GetTupleSize() Then
                Return False
            End If
            Dim i As Long
            Dim cnt As Long
            cnt = 0
            For i = 0 To GetTupleSize() - 1
                If (val(i) <> t1.GetC(i)) Then
                    cnt = cnt + 1
                    If cnt >= 2 Then
                        Return False
                    End If
                End If
            Next
            Return True
        End Function

        ' 和の公式を適用する
        Public Function SimpleOr(ByVal t1 As T) As T
            Dim ans() As Long
            ReDim ans(GetTupleSize() - 1)
            Dim i As Long
            For i = 0 To GetTupleSize() - 1
                ans(i) = GetC(i) Or t1.GetC(i)
            Next
            Return New T(ans)
        End Function

        ' 和の公式 No.2 が適用可能かをチェックする
        Public Function CanOrFormula2(ByVal t1 As T) As Long
            If t1.GetTupleSize() <> GetTupleSize() Then
                Return False
            End If
            Dim i As Long
            Dim cnt As Long
            Dim k As Long
            cnt = 0
            For i = 0 To GetTupleSize() - 1
                Dim a = GetC(i)
                Dim b = t1.GetC(i)
                If (a And b) > 0 AndAlso (a Or b) = b Then
                    ' OK
                Else
                    k = i
                    cnt = cnt + 1
                    If cnt >= 2 Then
                        Return -1
                    End If
                End If
            Next
            If cnt = 0 Then
                For i = 0 To GetTupleSize() - 1
                    If GetC(i) <> t1.GetC(i) Then
                        k = i
                        Exit For
                    End If
                Next
            End If
            Return k
        End Function

        ' 和の公式 No.2 を適用する
        Public Function OrFormula2(ByVal k As Long, ByVal t1 As T) As T
            Dim ret = Copy()
            ret.val(k) = GetC(k) Or t1.GetC(k)
            Return ret
        End Function

        ' ゼロかを判定する
        Public Function IsZero() As Boolean
            Dim i As Long
            For i = 0 To GetTupleSize() - 1
                If (val(i) = 0) Then
                    Return True
                End If
            Next
            Return False
        End Function

        ' Ωかを判定する
        Public Function IsOne() As Boolean
            Dim i As Long
            For i = 0 To GetTupleSize() - 1
                If (val(i) <> 7) Then
                    Return False
                End If
            Next
            Return True
        End Function

        Private val As Long()
    End Class

    ' 複数項を表すクラス
    Public Class TS

        ' Tクラスの配列を指定する
        Public Sub New(ByVal ParamArray Values())

            ReDim val(Values.Length - 1)
            Dim index As Long
            For index = 0 To Values.Length - 1
                val(index) = Values(index)
            Next
        End Sub

        ' コピーを作成する
        Public Function Copy() As TS
            Dim val_copy() As T
            ReDim val_copy(Size() - 1)
            Dim index As Long
            For index = 0 To Size() - 1
                val_copy(index) = val(index).Copy()
            Next
            Return New TS(val_copy)
        End Function

        ' タプルの数を返す
        Public Function Size() As Long
            Return val.Length
        End Function

        ' 指定位置のTを返す
        Public Function GetT(ByVal index As Long) As T
            Return val(index)
        End Function

        ' Andを計算する
        Public Function AndTS(ByVal t1 As T) As TS

            Dim ans As T()
            ReDim ans(Size() - 1)

            Dim index As Long
            For index = 0 To Size() - 1
                ans(index) = GetT(index).AndT(t1)
            Next
            Return New TS(ans)
        End Function

        ' Andを計算する
        Public Function AndTS(ByVal ts1 As TS) As TS

            Dim index As Long

            Dim ans As TS = New TS()
            For index = 0 To ts1.Size - 1
                Dim ts_tmp As TS
                ts_tmp = AndTS(ts1.GetT(index)) ' 処理を任せている
                If index = 0 Then
                    ans = ts_tmp
                Else
                    ans.AddSimple(ts_tmp)
                End If
            Next
            Return ans
        End Function

        ' Orを計算する
        Public Function OrTS(ByVal ts1 As TS) As TS
            Dim ret As TS
            ret = Copy()
            ret.AddSimple(ts1)
            Return ret
        End Function

        ' NegTSを計算する
        Public Function NegTS() As TS
            Dim ret As TS = New TS()
            Dim index As Long
            For index = 0 To Size() - 1
                If index = 0 Then
                    ret = GetT(0).NegT()
                Else
                    ret = ret.AndTS(GetT(index).NegT())
                End If
            Next
            Return ret
        End Function

        ' 単純に要素を追加する.．
        Public Sub AddSimple(ByVal t1 As T)

            ' サイズチェック
            If (val.Length > 0 AndAlso val(0).GetTupleSize <> t1.GetTupleSize()) Then
                Throw New ApplicationException("項の数の違うタプルを計算しました")
            End If

            ' 同一要素が存在しないかのチェック
            Dim index As Long
            For index = 0 To val.Length - 1
                If t1 = val(index) Then
                    Return
                End If
            Next

            ' 追加
            ReDim Preserve val(val.Length)
            val(val.Length - 1) = t1
        End Sub

        ' 単純に要素を追加する．
        Public Sub AddSimple(ByVal ts1 As TS)

            Dim index As Long
            For index = 0 To ts1.Size - 1
                AddSimple(ts1.GetT(index))
            Next
        End Sub

        ' 文字列出力
        Public Overrides Function ToString() As String
            Dim str As String
            str = ""
            Dim index As Long
            For index = 0 To val.Length - 1
                If index > 0 Then
                    str += "+" ' 最初以外に+を追加
                End If
                str += val(index).ToString
            Next
            Return str
        End Function

        ' 文字列出力(折り返し付)
        Public Function ToString2() As String
            Dim str As String
            str = ""
            Dim index As Long
            For index = 0 To val.Length - 1
                If index > 0 Then
                    str += "+" ' 最初以外に+を追加
                    str += vbCrLf
                End If
                str += val(index).ToString
            Next
            Return str
        End Function

        Public Sub DeleteT(ByVal index As Long)
            Dim j As Long
            For j = index To val.Length - 2
                val(j) = val(j + 1)
            Next j
            ReDim Preserve val(val.Length - 2)
        End Sub

        ' 縮約計算
        Public Function Reduce() As TS

            Dim i As Long, j As Long
            Dim ans As TS
            ans = Copy()

            ' 000化
            For i = 0 To ans.val.Length - 1
                If i >= ans.val.Length Then
                    Exit For
                End If
                If ans.val(i).IsZero() Then
                    If ans.val.Length > 1 Then
                        ans.DeleteT(i)
                        i = i - 1
                    Else
                        'ans.val(0) = ans.val(0).Zero()  ' 解を(000,000,000)のような形にするときに必要
                    End If
                End If
            Next

            ' 和の公式
            Dim loopflag As Boolean
            loopflag = True
            While loopflag
                loopflag = False
                For i = 0 To ans.val.Length - 1
                    If i >= ans.val.Length Then
                        Exit For
                    End If
                    For j = i + 1 To ans.val.Length - 1
                        If j >= ans.val.Length Then
                            Exit For
                        End If
                        If ans.val(i).CanSimpleOr(ans.val(j)) Then
                            ans.val(i) = ans.val(i).SimpleOr(ans.val(j))
                            ans.DeleteT(j)
                            loopflag = True
                        End If
                    Next
                Next
            End While

            ' Φ化
            For i = 0 To ans.val.Length - 1
                If ans.val(i).IsOne() Then
                    ReDim Preserve ans.val(0)
                    ans.val(0) = ans.val(0).One()
                    Exit For
                End If
            Next
            Return ans
        End Function

        ' 和の公式 No.2
        Public Function Reduce2() As TS

            Dim ans As TS
            ans = Copy()

            ' 和の公式
            Dim i As Long, j As Long
            Dim loopflag As Boolean
            loopflag = True
            While loopflag
                loopflag = False ' 変化がなくなるまで繰り返す

                ' 和の拡張公式(1)
                ' A∩B=Bならば，A∪B=Aとなる
                For i = 0 To ans.val.Length - 1
                    For j = i + 1 To ans.val.Length - 1
                        If i >= ans.val.Length Then Exit For
                        If j >= ans.val.Length Then Exit For
                        If ans.val(i).AndT(ans.val(j)) = ans.val(j) Then
                            ans.DeleteT(j)
                            loopflag = True
                            j = j - 1
                        ElseIf ans.val(i).AndT(ans.val(j)) = ans.val(i) Then
                            ans.DeleteT(i)
                            loopflag = True
                            If i > 0 Then
                                i = i - 1
                            End If
                            Exit For
                        End If
                    Next
                Next

                ' 和の拡張公式(2)
                For i = 0 To ans.val.Length - 1
                    If i >= ans.val.Length Then
                        Exit For
                    End If
                    For j = 0 To ans.val.Length - 1
                        If i = j Then Continue For
                        If j >= ans.val.Length Then
                            Exit For
                        End If
                        Dim k = ans.val(i).CanOrFormula2(ans.val(j))
                        If k >= 0 Then
                            Dim prev = ans.val(i)
                            ans.val(i) = ans.val(i).OrFormula2(k, ans.val(j))
                            If ans.val(i) <> prev Then
                                loopflag = True
                            End If
                        End If
                    Next
                Next

            End While
            Return ans
        End Function

        Private val() As T
    End Class
End Namespace

